# -*- coding: utf-8 -*-
from threading import Lock
from typing import Dict, Type


class SingletonMeta(type):
    _instances: Dict[Type, object] = dict()
    _lock = Lock()

    def __call__(cls, *args, **kwargs):
        with cls._lock:
            if cls not in cls._instances:
                instance = super().__call__(*args, **kwargs)
                cls._instances[cls] = instance
        return cls._instances[cls]



# -*- coding: utf-8 -*-
from typing import Union, List, Dict

import ujson

JsonType = Union[None, bool, int, float, str, List["JsonType"], Dict[str, "JsonType"]]


def fast_loads(str_json: str) -> JsonType:
    return ujson.loads(str_json)


def fast_dumps(obj: JsonType, ensure_ascii: bool = False, **kwargs) -> str:
    return ujson.dumps(obj, ensure_ascii=ensure_ascii, **kwargs)



# -*- coding: utf-8 -*-


def get_readable_size(size: float, round_number: int = 2) -> str:
    """
    Usage::
        >>> get_readable_size(0)
        '0B'
        >>> get_readable_size(1000)
        '1000B'
        >>> get_readable_size(1024)
        '1.0KB'
        >>> get_readable_size(2000)
        '1.95KB'
        >>> get_readable_size(1024 * 1024)
        '1.0MB'
        >>> get_readable_size(int(1024 * 1024 * 1.94))
        '1.94MB'
        >>> get_readable_size(1024 * 1024 * 1024 * 3)
        '3.0GB'
        >>> get_readable_size(1024 * 1024 * 1024 * 1024 * 4)
        '4.0TB'
        >>> get_readable_size(1024 * 1024 * 1024 * 1024 * 1024 * 5)
        '5.0PB'
        >>> get_readable_size(1024 * 1024 * 1024 * 1024 * 1024 * 1024 * 10)
        '10240.0PB'
    """
    unit_list = ["B", "KB", "MB", "GB", "TB", "PB"]
    index = 0

    while size >= 1024 and index < len(unit_list) - 1:
        size /= 1024
        index += 1

    return f"{round(size, round_number)}{unit_list[index]}"


def get_readable_duration(seconds: float, round_number: int = 2) -> str:
    """
    Usage::
        >>> get_readable_duration(0)
        '0s'
        >>> get_readable_duration(1)
        '1s'
        >>> get_readable_duration(60)
        '1.0m'
        >>> get_readable_duration(3600)
        '1.0h'
        >>> get_readable_duration(86400)
        '1.0d'
    """
    unit_info_list = [("s", 60), ("m", 60), ("h", 24), ("d", 0)]
    duration = seconds
    index = 0
    while True:
        unit, threshold = unit_info_list[index]
        if duration < threshold or threshold == 0:
            return f"{round(duration, round_number)}{unit}"
        else:
            duration /= threshold
            index += 1



# -*- coding: utf-8 -*-
import inspect
from importlib import import_module
from types import ModuleType
from typing import List, Type, TypeVar, Any

T = TypeVar("T")


def find_all_subclasses(module: str | ModuleType, base_class: Type[T]) -> List[Type[T]]:
    if isinstance(module, str):
        module = import_module(module)

    all_class_list = list()
    for name in dir(module):
        obj = getattr(module, name)
        if inspect.isclass(obj):
            all_class_list.append(obj)

    all_subclass_list = list()
    for the_class in all_class_list:
        if issubclass(the_class, base_class) and the_class != base_class:
            all_subclass_list.append(the_class)

    return all_subclass_list


def import_string(dotted_path: str) -> Any:
    try:
        module_path, class_name = dotted_path.rsplit(".", 1)
    except ValueError as err:
        raise ImportError(f"{dotted_path} doesn't look like a module path") from err

    module = import_module(module_path)

    try:
        return getattr(module, class_name)
    except AttributeError as err:
        raise ImportError(f'Module "{module}" does not define a "{class_name}" attribute/class') from err
